# coding=utf-8
"""Tests that sync `OSTree`_ `repositories`_.

.. _OSTree:
    http://docs.pulpproject.org/plugins/pulp_ostree/
.. _repositories:
   http://docs.pulpproject.org/en/latest/dev-guide/integration/rest-api/repo/cud.html
"""
import unittest
from urllib.parse import urljoin

from pulp_smash import api, selectors, utils
from pulp_smash.constants import OSTREE_FEED, OSTREE_BRANCHES
from pulp_smash.pulp2.constants import REPOSITORY_PATH
from pulp_smash.pulp2.utils import BaseAPITestCase, sync_repo

from pulp_2_tests.tests.ostree.utils import gen_repo
from pulp_2_tests.tests.ostree.utils import set_up_module as setUpModule  # pylint:disable=unused-import


def _sync_repo(cfg, href):
    """Sync a repository and wait for the sync to complete.

    Verify only the HTTP status codes of Pulp's responses. Don't verify
    response contents, as the default response handler does. Return ``call
    report, tasks``.
    """
    response = api.Client(cfg, api.code_handler).post(
        urljoin(href, 'actions/sync/'),
        {'override_config': {}},
    )
    tasks = tuple(api.poll_spawned_tasks(cfg, response.json()))
    return response, tasks


# It's OK for a mixin to have just one method.
class _SyncMixin():  # pylint:disable=too-few-public-methods
    """Add test methods verifying Pulp's behaviour when a sync is started.

    The ``report`` instance attribute should be available. This is the response
    to a "sync repository" API call: a call report.
    """

    def test_start_sync_code(self):
        """Assert the call to sync a repository returns an HTTP 202."""
        self.assertEqual(self.report.status_code, 202)


class _SyncFailedMixin():
    """Add test methods verifying Pulp's behaviour when a sync fails.

    This mixin assumes that:

    * The ``report`` and ``tasks`` instance attributes are available. The
      former should be the response to a "sync repository" API call — a call
      report. The latter should be a collection of completed task states.
    * The child class making use of this mixin is a unittest-compatible test
      case.
    """

    def test_number_tasks(self):
        """Assert that exactly one task was spawned."""
        self.assertEqual(len(self.tasks), 1, self.tasks)

    def test_task_error_traceback(self):
        """Assert each task's "error" and "traceback" fields are non-null."""
        for i, task in enumerate(self.tasks):
            for key in {'error', 'traceback'}:
                with self.subTest((i, key)):
                    self.assertIsNotNone(task[key])


# It's OK for a mixin to have just one method.
class _SyncImportFailedMixin():  # pylint:disable=too-few-public-methods
    """Add test methods verifying Pulp's behaviour when a sync fails.

    This class is like ``_SyncFailedMixin``, with the additional restriction
    that the repo being synced is an OSTree repo with a feed and/or branches.
    """

    def test_error_details(self):
        """Assert that each task's progress report contains error details."""
        for task in self.tasks:
            num_error_details = sum(
                1 if action['error_details'] != [] else 0
                for action in task['progress_report']['ostree_web_importer']
            )
            self.assertGreater(num_error_details, 0)


class SyncTestCase(_SyncMixin, BaseAPITestCase):
    """Create an OSTree repository with a valid feed and branch, and sync it.

    The sync should complete with no errors reported.
    """

    @classmethod
    def setUpClass(cls):
        """Create an OSTree repository with a valid feed and branch."""
        super().setUpClass()
        if not selectors.bug_is_fixed(1934, cls.cfg.pulp_version):
            raise unittest.SkipTest('https://pulp.plan.io/issues/1934')
        body = gen_repo()
        body['importer_config']['feed'] = OSTREE_FEED
        body['importer_config']['branches'] = OSTREE_BRANCHES
        repo = api.Client(cls.cfg).post(REPOSITORY_PATH, body).json()
        cls.resources.add(repo['_href'])
        cls.report = sync_repo(cls.cfg, repo)
        cls.tasks = tuple(api.poll_spawned_tasks(cls.cfg, cls.report.json()))

    def test_task_progress_report(self):
        """Assert no task's progress report contains error details."""
        for task in self.tasks:
            for action in task['progress_report']['ostree_web_importer']:
                with self.subTest(task=task):
                    self.assertEqual(action['error_details'], [])


class SyncInvalidFeedTestCase(
        _SyncMixin,
        _SyncFailedMixin,
        _SyncImportFailedMixin,
        BaseAPITestCase):
    """Create an OSTree repository with an invalid feed and sync it."""

    @classmethod
    def setUpClass(cls):
        """Set ``cls.body``."""
        super().setUpClass()
        client = api.Client(cls.cfg)
        body = gen_repo()
        body['importer_config']['feed'] = utils.uuid4()
        body['importer_config']['branches'] = OSTREE_BRANCHES
        repo_href = client.post(REPOSITORY_PATH, body).json()['_href']
        cls.resources.add(repo_href)
        cls.report, cls.tasks = _sync_repo(cls.cfg, repo_href)


class SyncInvalidBranchesTestCase(
        _SyncMixin,
        _SyncFailedMixin,
        _SyncImportFailedMixin,
        BaseAPITestCase):
    """Create an OSTree repository with invalid branches and sync it."""

    @classmethod
    def setUpClass(cls):
        """Create and sync an OSTree repository."""
        super().setUpClass()
        client = api.Client(cls.cfg)
        body = gen_repo()
        body['importer_config']['feed'] = OSTREE_FEED
        body['importer_config']['branches'] = [utils.uuid4()]
        repo_href = client.post(REPOSITORY_PATH, body).json()['_href']
        cls.resources.add(repo_href)
        cls.report, cls.tasks = _sync_repo(cls.cfg, repo_href)


class SyncMissingAttrsTestCase(
        _SyncMixin,
        _SyncFailedMixin,
        BaseAPITestCase):
    """Create an OSTree repository with no feed or branches and sync it."""

    @classmethod
    def setUpClass(cls):
        """Create and sync an OSTree repository."""
        super().setUpClass()
        client = api.Client(cls.cfg)
        body = gen_repo()
        repo_href = client.post(REPOSITORY_PATH, body).json()['_href']
        cls.resources.add(repo_href)
        cls.report, cls.tasks = _sync_repo(cls.cfg, repo_href)
