# coding=utf-8
"""Tests for the `remove_missing`_ repository option.

.. _remove_missing:
    https://docs.pulpproject.org/plugins/pulp_rpm/tech-reference/yum-plugins.html
"""
import random
import unittest
from urllib.parse import urljoin

from pulp_smash import api, config, selectors
from pulp_smash.constants import RPM_SIGNED_FEED_URL, RPM_UNSIGNED_FEED_URL
from pulp_smash.pulp2.constants import ORPHANS_PATH, REPOSITORY_PATH
from pulp_smash.pulp2.utils import publish_repo, search_units, sync_repo

from pulp_2_tests.tests.rpm.api_v2.utils import gen_distributor, gen_repo
from pulp_2_tests.tests.rpm.utils import (
    check_issue_2277,
    check_issue_2620,
    check_issue_3104,
)
from pulp_2_tests.tests.rpm.utils import set_up_module as setUpModule  # pylint:disable=unused-import

_PUBLISH_DIR = 'pulp/repos/'


class RemoveMissingTestCase(unittest.TestCase):
    """Test the `remove_missing`_ repository option.

    The test procedure is as follows:

    1. Create a repository, populate it with content, and publish it.
    2. Create several more repositories whose "feed" attribute references the
       repository created in the previous step. Sync each of these child
       repositories.
    3. Remove a content unit from the parent repository and re-publish it.
    4. Re-sync each of the child repositories.

    The child repositories are designed to test the following issues:

    * `Pulp #1621 <https://pulp.plan.io/issues/1621>`_
    * `Pulp #2503 <https://pulp.plan.io/issues/2503>`_

    .. _remove_missing:
        https://docs.pulpproject.org/plugins/pulp_rpm/tech-reference/yum-plugins.html
    """

    @classmethod
    def setUpClass(cls):
        """Initialize class-wide variables."""
        cls.cfg = config.get_config()
        cls.repos = {}  # Each inner dict has info about a repository.
        if check_issue_3104(cls.cfg):
            raise unittest.SkipTest('https://pulp.plan.io/issues/3104')
        if check_issue_2277(cls.cfg):
            raise unittest.SkipTest('https://pulp.plan.io/issues/2277')
        if check_issue_2620(cls.cfg):
            raise unittest.SkipTest('https://pulp.plan.io/issues/2620')

    @classmethod
    def tearDownClass(cls):
        """Delete all resources named by ``resources``."""
        client = api.Client(cls.cfg)
        for repo in cls.repos.values():
            client.delete(repo['_href'])
        client.delete(ORPHANS_PATH)

    def test_01_create_root_repo(self):
        """Create, sync and publish a repository.

        The repositories created in later steps sync from this one.
        """
        client = api.Client(self.cfg, api.json_handler)
        body = gen_repo()
        body['importer_config']['feed'] = RPM_SIGNED_FEED_URL
        body['distributors'] = [gen_distributor()]
        self.repos['root'] = client.post(REPOSITORY_PATH, body)
        self.repos['root'] = _get_details(self.cfg, self.repos['root'])
        sync_repo(self.cfg, self.repos['root'])
        publish_repo(self.cfg, self.repos['root'])

    def test_02_create_immediate_child(self):
        """Create a child repository with the "immediate" download policy.

        Sync the child repository, and verify it has the same contents as the
        root repository.
        """
        # We disable SSL validation for a practical reason: each HTTPS feed
        # must have a certificate to work, which is burdensome to do here.
        client = api.Client(self.cfg, api.json_handler)
        body = gen_repo()
        body['importer_config']['feed'] = urljoin(
            self.cfg.get_base_url(),
            _PUBLISH_DIR +
            self.repos['root']['distributors'][0]['config']['relative_url'],
        )
        body['importer_config']['remove_missing'] = True
        body['importer_config']['ssl_validation'] = False
        self.repos['immediate'] = client.post(REPOSITORY_PATH, body)
        self.repos['immediate'] = (
            _get_details(self.cfg, self.repos['immediate'])
        )
        sync_repo(self.cfg, self.repos['immediate'])

        # Verify the two repositories have the same contents.
        root_ids = _get_rpm_ids(_get_rpms(self.cfg, self.repos['root']))
        immediate_ids = _get_rpm_ids(
            _get_rpms(self.cfg, self.repos['immediate'])
        )
        self.assertEqual(root_ids, immediate_ids)

    def test_02_create_on_demand_child(self):
        """Create a child repository with the "on_demand" download policy.

        Also, let the repository's "remove_missing" option be true. Then, sync
        the child repository, and verify it has the same contents as the root
        repository.
        """
        client = api.Client(self.cfg, api.json_handler)
        body = gen_repo()
        body['importer_config']['feed'] = urljoin(
            self.cfg.get_base_url(),
            _PUBLISH_DIR +
            self.repos['root']['distributors'][0]['config']['relative_url'],
        )
        body['importer_config']['download_policy'] = 'on_demand'
        body['importer_config']['remove_missing'] = True
        body['importer_config']['ssl_validation'] = False
        self.repos['on demand'] = client.post(REPOSITORY_PATH, body)
        self.repos['on demand'] = (
            _get_details(self.cfg, self.repos['on demand'])
        )
        sync_repo(self.cfg, self.repos['on demand'])

        # Verify the two repositories have the same contents.
        root_ids = _get_rpm_ids(_get_rpms(self.cfg, self.repos['root']))
        on_demand_ids = _get_rpm_ids(
            _get_rpms(self.cfg, self.repos['on demand'])
        )
        self.assertEqual(root_ids, on_demand_ids)

    def test_03_update_root_repo(self):
        """Remove a content unit from the root repository and republish it."""
        unit = random.choice(_get_rpms(self.cfg, self.repos['root']))
        criteria = {
            'filters': {'unit': {'name': unit['metadata']['name']}},
            'type_ids': [unit['unit_type_id']],
        }
        api.Client(self.cfg).post(
            urljoin(self.repos['root']['_href'], 'actions/unassociate/'),
            {'criteria': criteria}
        ).json()
        publish_repo(self.cfg, self.repos['root'])

        # Verify the removed unit cannot be found via a search.
        units = search_units(self.cfg, self.repos['root'], criteria)
        self.assertEqual(len(units), 0, units)

    def test_04_sync_immediate_child(self):
        """Sync the "immediate" repository.

        Verify it has the same contents as the root repository.
        """
        sync_repo(self.cfg, self.repos['immediate'])
        root_ids = _get_rpm_ids(_get_rpms(self.cfg, self.repos['root']))
        immediate_ids = _get_rpm_ids(
            _get_rpms(self.cfg, self.repos['immediate'])
        )
        self.assertEqual(root_ids, immediate_ids)

    def test_04_sync_on_demand_child(self):
        """Sync the "on demand" repository.

        Verify it has the same contents as the root repository.
        """
        sync_repo(self.cfg, self.repos['on demand'])
        root_ids = _get_rpm_ids(_get_rpms(self.cfg, self.repos['root']))
        on_demand_ids = _get_rpm_ids(
            _get_rpms(self.cfg, self.repos['on demand'])
        )
        self.assertEqual(root_ids, on_demand_ids)


class RemoveCountTestCase(unittest.TestCase):
    """Re-sync a child repository with the ``remove_missing`` enabled.

    After removing an arbitrary number of units from the parent repository.

    Do the following:

    1. Create 1st repository with, feed, sync and publish it.
    2. Create 2nd repository with feed pointed to 1st repository, and
       ``remove_missing`` enabled, sync.
    3. Remove an arbitrary number of units from 1st repository, re-publish it.
    4. Re-sync the 2nd repository. Assert that ``removed_count`` is equal to
       number of removed units.

    `Pulp issue #2616 <https://pulp.plan.io/issues/2616>`_ caused after sync
    with remove_missing option the next publish is no-op.
    """

    def test_all(self):
        """Re-sync a child repository with the ``remove_missing`` enabled."""
        repos = []
        cfg = config.get_config()
        if not selectors.bug_is_fixed(2616, cfg.pulp_version):
            self.skipTest('https://pulp.plan.io/issues/2616')

        # Create 1st repo, sync and publish it.
        client = api.Client(cfg, api.json_handler)
        body = gen_repo()
        body['importer_config']['feed'] = RPM_UNSIGNED_FEED_URL
        body['distributors'] = [gen_distributor()]
        repos.append(client.post(REPOSITORY_PATH, body))
        self.addCleanup(client.delete, repos[0]['_href'])
        repos[0] = _get_details(cfg, repos[0])
        sync_repo(cfg, repos[0])
        publish_repo(cfg, repos[0])

        # Create 2nd repo, sync.
        body = gen_repo()
        body['importer_config']['feed'] = urljoin(
            cfg.get_base_url(),
            _PUBLISH_DIR +
            repos[0]['distributors'][0]['config']['relative_url'],
        )
        body['importer_config']['remove_missing'] = True
        repos.append(client.post(REPOSITORY_PATH, body))
        self.addCleanup(client.delete, repos[1]['_href'])
        repos[1] = _get_details(cfg, repos[1])
        sync_repo(cfg, repos[1])

        # Remove an arbitrary number of units from 1st repo, re-publish it.
        units = _get_rpms(cfg, repos[0])
        marked_units = random.sample(units, random.randint(1, len(units)))
        for marked_unit in marked_units:
            criteria = {
                'filters': {'unit': {'name': marked_unit['metadata']['name']}},
                'type_ids': [marked_unit['unit_type_id']],
            }
            client.post(
                urljoin(repos[0]['_href'], 'actions/unassociate/'),
                {'criteria': criteria},
            )
        publish_repo(cfg, repos[0])

        # Re-sync 2nd repo.
        report = sync_repo(cfg, repos[1])
        tasks = tuple(api.poll_spawned_tasks(cfg, report.json()))
        self.assertEqual(
            tasks[0]['result']['removed_count'],
            len(marked_units),
        )


def _get_rpms(cfg, repo):
    """Return RPM content units in the given repository."""
    return search_units(cfg, repo, {'type_ids': ('rpm',)})


def _get_rpm_ids(search_body):
    """Get RPM unit IDs from repository search results. Return as a set."""
    return {
        unit['unit_id'] for unit in search_body
        if unit['unit_type_id'] == 'rpm'
    }


def _get_details(cfg, repo):
    """Get detailed information about the given repository."""
    return api.Client(cfg).get(repo['_href'], params={'details': True}).json()
